package BAMMatrix;
# "theoretically" simpler version of GenotypeMatrix.pm
# build a genotype matrix of reads in a BAM file

use strict;
use Carp qw(confess);

use File::Basename;
use Digest::MD5 qw(md5_hex);

use Configurable;
use JavaRun;
use Cluster;
use JobQueue;
use SJConstants;
use AtomicOutfile;
use Reporter;
use DelimitedFile;
use Counter;

#use misc;
use FileUtils qw(newer_than);

use SampleName;

@BAMMatrix::ISA = qw(Configurable Exporter);
@BAMMatrix::EXPORT_OK = qw();

use MethodMaker qw(
snvs
bams
job_id
final_matrix
matrix_delimiter
matrix_format

bam_label_basenames
bam_label_sample
cell_delimiter

cell_format_variant_then_reference
variant_label
annotation_callback
extra_fields
warn
fraction_mode

variant_count_first
java_ram

snv_filter_bam
snv_filter_min_variant_reads

verbose

idataplex

step_mode
step1_outfile

working_dir
min_quality
		  );

sub new {
  my ($type, %options) = @_;
  my $self = {};
  bless $self, $type;
  $self->bam_label_basenames(1);
  $self->cell_format_variant_then_reference(1);
  $self->matrix_delimiter("\t");
  $self->cell_delimiter("|");
  $self->bams([]);
  $self->snvs({});
  $self->matrix_format("sample_y");
  $self->variant_label("variant");
  $self->variant_count_first(1);
#  $self->verbose(1);

#  $self->java_ram(2000);
  # occasional killoff:
  # TERM_SWAP: job killed after reaching LSF swap usage limit.

#  $self->java_ram(1500);
  $self->java_ram(500);

  $self->configure(%options);
  return $self;
}

sub add_snv {
  # add a single SNV to the target list
  my ($self, $ref_name, $position, $base_ref, $base_var) = @_;
  foreach (@_) {
    die unless defined $_;
  }
  die unless @_ == 5;
  confess "ERROR: possible non-SNV base detected: ref=$base_ref var=$base_var" unless length($base_ref) == 1 and length($base_var) == 1;

  $self->snvs()->{$ref_name}{$position}{$base_ref}{$base_var} = 1;
}

sub add_bam {
  my ($self, $bam) = @_;
  confess sprintf "%s does not exist or is zero-length", $bam unless -s $bam;
  my $bai = $bam . ".bai";
  confess sprintf "%s does not have an index file", $bam, $bai unless -s $bai;
  push @{$self->bams()}, $bam;
}

sub generate_matrix {
  my ($self) = @_;
  $self->build_query_jobs();
  # real-time
  $self->run_query_jobs();
  # clustered

  my $bams = $self->bams();
  die "no BAMs!" unless @{$bams};

  my %saw;
  my $unique = 1;
  foreach my $bam (@{$bams}) {
    my $bn = basename($bam);
    if ($saw{$bn}) {
      $unique = 0;
    } else {
      $saw{$bn} = 1;
    }
  }
  $self->bam_label_basenames(0) unless $unique;
  # if we are comparing BAMs with the same name in different
  # directories, basename label is insufficient

  my @outfiles = map {$self->get_sites_results($_)} @{$bams};
#  die scalar @outfiles;

  if (my $mode = $self->step_mode()) {
    $self->write_results() if $mode == 2;
  } else {
    my $jq = new JobQueue();
    foreach (@outfiles) {
      printf STDERR "outfile expected: %s\n", $_;
    }

    $jq->add_level("-callback" => 0,
		   "-outfiles" => \@outfiles);
    $jq->add_level("-callback" => [$self, "write_results"],
		   "-wait" => 0);
    $jq->run();
  }
}

sub write_results {
  my ($self, %options) = @_;
  my $delimiter = $self->matrix_delimiter() || die;

  my $matrix_format = $self->matrix_format || die;

  my $matrix = $self->get_matrix();
  my $outfile = $self->get_matrix_outfile() || die;

  my @labels = "bam";
  my @ordered_labels;

  my $bams = $self->bams();

  my $variant_label = $self->variant_label();
  my $annotation_callback = $self->annotation_callback();
  my $verbose = $self->verbose();
  printf STDERR "generating final matrix...\n" if $verbose;

  if ($matrix_format eq "sample_y") {
    #
    # samples on the Y axis (rows)
    #
    my %by_bam;

    my $c = new Counter(scalar keys %{$matrix});
    foreach my $chr (sort keys %{$matrix}) {
      foreach my $pos_hg19 (sort {$a <=> $b} keys %{$matrix->{$chr}}) {
	foreach my $ref_base (sort keys %{$matrix->{$chr}{$pos_hg19}}) {
	  foreach my $var_base (sort keys %{$matrix->{$chr}{$pos_hg19}{$ref_base}}) {
	    my $label = join ".", $chr, $pos_hg19, $ref_base, $var_base;
	    push @ordered_labels, $label;
	    foreach my $bam (@{$bams}) {
	      my $cell = $matrix->{$chr}{$pos_hg19}{$ref_base}{$var_base}{$bam} || die "no data for $chr $pos_hg19 $ref_base $var_base $bam";
	      $by_bam{$bam}{$label} = $self->format_matrix_cell($cell);
	    }
	  }
	}
      }
      $c->next($chr) if $verbose;
    }

    my $rpt = new Reporter(
			   "-delimiter" => $delimiter,
			   "-file" => $outfile,
			   "-labels" => [
			     "bam",
			     @ordered_labels
      ]
	);
    foreach my $bam (@{$bams}) {
      my %r;
      $r{bam} = $self->get_bam_label($bam);
      foreach my $label (@ordered_labels) {
	$r{$label} = $by_bam{$bam}{$label};
      }
#      &$annotation_callback(\%r) if $annotation_callback;
      $rpt->end_row(\%r);
    }
    $rpt->finish();
  } elsif ($matrix_format eq "sample_x") {
    #
    #  samples on the X axis (columns)
    #
    my @l;
    push @l, $variant_label;
    push @l, @{$self->extra_fields} if $self->extra_fields();
    foreach my $bam (@{$bams}) {
      push @l, $self->get_bam_label($bam);
    }

    my $rpt = new Reporter(
			   "-delimiter" => $delimiter,
			   "-file" => $outfile,
			   "-labels" => \@l
			  );

    foreach my $chr (sort keys %{$matrix}) {
      foreach my $pos_hg19 (sort {$a <=> $b} keys %{$matrix->{$chr}}) {
	foreach my $ref_base (sort keys %{$matrix->{$chr}{$pos_hg19}}) {
	  foreach my $var_base (sort keys %{$matrix->{$chr}{$pos_hg19}{$ref_base}}) {
	    my $label = join ".", $chr, $pos_hg19, $ref_base, $var_base;

	    my %r;
	    $r{$variant_label} = $label;
	    foreach my $bam (@{$bams}) {
	      my $cell = $matrix->{$chr}{$pos_hg19}{$ref_base}{$var_base}{$bam} || die "no data for $chr $pos_hg19 $ref_base $var_base $bam";
	      $r{$self->get_bam_label($bam)} = $self->format_matrix_cell($cell);
	    }
	    &$annotation_callback(\%r) if $annotation_callback;
	    $rpt->end_row(\%r);
	  }
	}
      }
    }

    $rpt->finish();
  } elsif ($matrix_format eq "simple") {
    # simple non-normalized report

    my @l = (
	     "bam",
	     $variant_label,
	    );
    push @l, @{$self->extra_fields} if $self->extra_fields();
    push @l, (
	      "reference_count",
	      "variant_count",
	     );

    my $rpt = new Reporter(
			   "-delimiter" => $delimiter,
			   "-file" => $outfile,
			   "-labels" => \@l,
			  );

    my $cell_delimiter = $self->cell_delimiter;
    $cell_delimiter = "\\" . $cell_delimiter if $cell_delimiter eq "|";

    foreach my $chr (sort keys %{$matrix}) {
      foreach my $pos_hg19 (sort {$a <=> $b} keys %{$matrix->{$chr}}) {
	foreach my $ref_base (sort keys %{$matrix->{$chr}{$pos_hg19}}) {
	  foreach my $var_base (sort keys %{$matrix->{$chr}{$pos_hg19}{$ref_base}}) {
	    my $label = join ".", $chr, $pos_hg19, $ref_base, $var_base;

	    my %r;
	    $r{$variant_label} = $label;
	    foreach my $bam (@{$bams}) {
	      $r{bam} = $self->get_bam_label($bam);
	      my $cell = $matrix->{$chr}{$pos_hg19}{$ref_base}{$var_base}{$bam};
	      die "no data for $chr $pos_hg19 $ref_base $var_base $bam" unless defined $cell;
	      my (@f) = split $cell_delimiter, $cell;
	      die "format fail for $cell" unless @f == 2;
	      @r{qw(variant_count reference_count)} = @f;

	      &$annotation_callback(\%r) if $annotation_callback;
	      $rpt->end_row(\%r);
	    }
	  }
	}
      }
    }
    $rpt->finish();
  } else {
    die "unknown matrix format $matrix_format";
  }
}


sub get_matrix {
  #
  # build final integrated data matrix
  #
  my ($self, %options) = @_;

  return $self->final_matrix() if $self->final_matrix();

  my $bam_info = $self->get_bam_results();
  # get low-level BAM observations

  my $snvs = $self->snvs();
  my $bams = $self->bams;

  my $snv_filter_bam = $self->snv_filter_bam();
  my $snv_filter_min_variant_reads = $self->snv_filter_min_variant_reads();

  my %matrix;

  my $cell_delim = $self->cell_delimiter || die;

  print STDERR "building matrix...\n";
  my $c = new Counter(scalar keys %{$snvs});

  my $fraction_mode = $self->fraction_mode();

  foreach my $chr (keys %{$snvs}) {
    foreach my $pos (keys %{$snvs->{$chr}}) {

      if (keys %{$snvs->{$chr}{$pos}} > 1) {
	my $msg = sprintf "ERROR: QC fail: multiple reference bases at %s.%d: %s", $chr, $pos, join ",", keys %{$snvs->{$chr}{$pos}};
	if ($self->warn) {
	  printf STDERR "%s\n", $msg;
	} else {
	  die $msg;
	}
      }

      foreach my $ref_base (keys %{$snvs->{$chr}{$pos}}) {
	foreach my $variant_base (keys %{$snvs->{$chr}{$pos}{$ref_base}}) {
	  my $usable = 1;
	  if ($snv_filter_bam) {
	    if ($snv_filter_min_variant_reads) {
	      my $variant_count = $bam_info->{$snv_filter_bam}{$chr}{$pos}{$variant_base};
	      if ($variant_count < $snv_filter_min_variant_reads) {
		printf STDERR "excluding %s: %d variant reads\n",
		join(".", $chr, $pos, $ref_base, $variant_base), $variant_count;
		$usable = 0;
	      }
	    } else {
	      die "unknown filtering";
	    }
	  }

	  if ($usable) {
	    foreach my $bam (@{$bams}) {
	      my $variant_count;
	      my $reference_count;

	      $reference_count = $bam_info->{$bam}{$chr}{$pos}{$ref_base};
	      $variant_count = $bam_info->{$bam}{$chr}{$pos}{$variant_base};
	      die "no data for ref count for $bam $chr $pos $ref_base!" unless defined $reference_count;
	      die "no data for variant count for $bam $chr $pos $variant_base!" unless defined $variant_count;
	      # can happen if cache is invalid (SNV list has changed)
	      
	      my $cell = join $cell_delim, $variant_count, $reference_count;
	      $matrix{$chr}{$pos}{$ref_base}{$variant_base}{$bam} = $cell;
	    }
	  }
	}
      }
    }
    $c->next($chr);

  }

  $self->final_matrix(\%matrix);
  return \%matrix;
}


sub get_bam_results {
  #
  # get data observed from BAMs
  #
  my ($self, %options) = @_;

  my %info;
  my $bams = $self->bams();
  my $verbose = $self->verbose();
  printf STDERR "loading BAM read count data...\n" if $verbose;

  my $c = new Counter($bams);
  foreach my $bam (@{$bams}) {
    my $outfile = $self->get_sites_results($bam);
    die "no data for $outfile" unless -s $outfile;
    printf STDERR "reading %s\n", $outfile if $outfile;
    my $df = new DelimitedFile(
      "-file" => $outfile,
      "-headers" => 1,
	);
    while (my $row = $df->get_hash()) {
      my $ref = $row->{reference} || die;
      my $bn = $row->{base_num} || die;
      foreach (qw(A C G T)) {
	die "no col for $_" unless exists $row->{$_};
      }
      die "duplicate" if $info{$bam}{$ref}{$bn};
      $info{$bam}{$ref}{$bn} = $row;
    }
    $c->next($bam) if $verbose;
  }
  return \%info;
}


sub build_query_jobs {
  #
  #  build job lists to query low-level BAM counts.
  #
  my ($self, %options) = @_;
  my $snvs = $self->snvs();
  my $bams = $self->bams();
  
  die "no SNVs specified" unless %{$snvs};

  foreach my $bam (@{$bams}) {
    my $outfile = $self->get_sites_file($bam);
    next if -s $outfile;
    my $rpt = new Reporter(
      "-delimiter" => "\t",
      "-file" => $outfile,
      "-labels" => [
	qw(
           reference
           base_num
          )
      ]);

    foreach my $chr (sort keys %{$snvs}) {
      foreach my $pos (sort {$a <=> $b} keys %{$snvs->{$chr}}) {
	my %r;
	$r{reference} = $chr;
	$r{base_num} = $pos;
	$rpt->end_row(\%r);
      }
    }
    $rpt->finish();
    die "where is $outfile" unless -s $outfile;
  }
}

sub run_query_jobs {
  my ($self) = @_;
  my $bams = $self->bams();

  my $java_ram = $self->java_ram();
  my $cluster_ram = $java_ram + SJConstants::JAVA_JOB_REQUEST_BUFFER_MB;

  my $step_mode = $self->step_mode();
  if ($step_mode) {
    my $outfile = $self->step1_outfile() || return;
    open(STEPTMP, ">" . $outfile) || die;
  }

  foreach my $bam (@{$bams}) {
    my $sites = $self->get_sites_file($bam);
    my $outfile = $self->get_sites_results($bam);

    unlink $outfile if -s $outfile and newer_than($bam, $outfile);
    # start over

    my $jr = new JavaRun("-classname" => "Ace2.ReadReport");
    $jr->ram($java_ram);
    my $cmd = sprintf '-min-mapq 0 -bam %s -passthrough %s', $bam, $sites;
    $cmd .= sprintf ' -min-quality %d', $self->min_quality if $self->min_quality;

    my $line = $jr->run("-command" => $cmd, "-return" => 1);
#    die $line;

    if ($self->step_mode) {
      printf STEPTMP "%s\n", $line;
    } else {
      my $c = new Cluster(
			  "-command" => $line,
			  "-outfile" => $outfile,
			  "-project" => "bam_genotype_matrix",
			  "-memory" => $cluster_ram,
			 );
      $c->node_class("idataplex") if $self->idataplex();
      $c->run();
    }
  }

  close STEPTMP if $step_mode;

}

sub get_cache_files {
  my ($self) = @_;
  my $job_id = $self->job_id || die "need job_id";
  my $dir = $self->working_dir() || ".";
  my @files = glob(sprintf "%s/sites_*", $dir, $job_id);
  return \@files;
}

sub get_sites_file {
  my ($self, $bam) = @_;
  my $job_id = $self->job_id || die "need job_id";
  my $dir = $self->working_dir() || ".";
  return sprintf "%s/sites_%s_%s_%s.tab",
    $dir, $job_id, basename($bam), md5_hex($bam);
  # add MD5 because basename may not be unique, i.e. if we are
  # comparing BAMs with the same name in different directories
  # TO DO: also add signature of variant list in case this is changed?
}

sub get_sites_results {
  my ($self, $bam) = @_;
  return $self->get_sites_file($bam) . ".read_report.tab";
}

sub get_matrix_outfile {
  my ($self) = @_;
  my $ao = new AtomicOutfile("-suffix" => "tab");
  $ao->base("matrix");
  $ao->add($self->job_id || die "need job_id");
  my $format = $self->matrix_format();
  $ao->add($format);
  unless ($format eq "simple") {
    $ao->add($self->cell_format_variant_then_reference() ? "cell_variant_count_first" : "cell_reference_count_first");
    # only add this element for actual matrix files
  }
  my $of = $ao->get_outfile;

  if (my $wd = $self->working_dir) {
    $of = sprintf '%s/%s', $wd, $of;
  }

  return $of;
}

sub get_bam_label {
  my ($self, $bam) = @_;
  if ($self->bam_label_sample()) {
    my %info = SampleName::parse(basename($bam));
    return $info{sample};
  } elsif ($self->bam_label_basenames()) {
    return basename($bam);
  } else {
    return $bam;
  }
}

sub format_matrix_cell {
  # optional formatting for matrix-style output
  my ($self, $cell) = @_;
  my $result = $cell;
  if ($self->fraction_mode()) {
    my ($variant_count, $reference_count);
    if ($self->variant_count_first()) {
      # this will change eventually
      my $cell_delimiter = $self->cell_delimiter;
      $cell_delimiter = "\\" . $cell_delimiter if $cell_delimiter eq "|";
      ($variant_count, $reference_count) = split $cell_delimiter, $cell;
    } else {
      die "FIX ME: only variant count first implemented";
    }
    my $total = $variant_count + $reference_count;
    if ($total == 0) {
      $result = "NaN";
      # TO DO: additional formatting options
    } else {
      $result = $variant_count / $total;
    }

  }
  return $result;
}

sub add_extra_field {
  my ($self, $field) = @_;
  my $ef = $self->extra_fields || [];
  $self->extra_fields($ef);
  unless (grep {$_ eq $field} @{$ef}) {
    # add only if not present already
    push @{$ef}, $field;
  }
}

sub cleanup {
  # remove intermediate and log files
  my ($self) = @_;

  my $bams = $self->bams();

  my @files;
  foreach my $bam (@{$bams}) {
    my $sites = $self->get_sites_file($bam);
    push @files, glob($sites . "*");
  }
  unlink @files;

}

1;
